/*!
 * Copyright (C) 2006 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log$
 */

/*!
 * \example testrtl/testrtl.c
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <string.h>

#include "rtlregs.h"

#define NIC_PAGE_SIZE   0x100
#define NIC_START_PAGE  0x40
#define NIC_STOP_PAGE   0x60
#define NIC_TX_PAGES    6
#define NIC_TX_BUFFERS      1
#define NIC_FIRST_TX_PAGE   NIC_START_PAGE
#define NIC_FIRST_RX_PAGE   (NIC_FIRST_TX_PAGE + NIC_TX_PAGES * NIC_TX_BUFFERS)

unsigned char my_mac[6] = { 0x00, 0x06, 0x98, 0xf0, 0x11, 0x23 };
unsigned char my_ip[4] = { 192, 168, 192, 254 };
unsigned char rem_ip[4] = { 192, 168, 192, 253 };

unsigned char arp_rqh[] = {
    0x08, 0x06, /* Ethernet type: ARP. */
    0x00, 0x01, /* Hardware type: Ethernet. */
    0x08, 0x00, /* Protocol type: IP. */
    0x06,       /* Hardware size: 6. */
    0x04,       /* Protocol size: 4. */
    0x00, 0x01  /* Opcode: Request. */
};

unsigned int base;
unsigned char packet[1500];

static int PutChar(char c, FILE *stream);

static FILE mystdout = FDEV_SETUP_STREAM(PutChar, NULL, _FDEV_SETUP_WRITE);

void Delay(unsigned long loops)
{
    while (loops--) {
        asm volatile("nop");
    }
}

static void StdIoInit(void)
{
    UCSR0B = _BV(RXEN) | _BV(TXEN);
    UBRR0L = 7;
    stdout = &mystdout;
    Delay(1000);
}

static int PutChar(char c, FILE *stream)
{
    if (c == '\n') {
        PutChar('\r', stream);
    }
    while((UCSR0A & _BV(UDRE)) == 0);
    UDR0 = c;

    return 0;
}

int AssembleArpRequest(unsigned char *dmac)
{
    int rc;

    memset(packet, 0, sizeof(packet));

    /* Destination MAC address. */
    if (dmac) {
        memcpy(packet, dmac, 6);
    }
    else {
        memset(packet, 0xFF, 6);
    }
    rc = 6;
    /* Source MAC address. */
    memcpy(&packet[rc], my_mac, 6);
    rc += 6;
    /* ARP request header. */
    memcpy(&packet[rc], arp_rqh, sizeof(arp_rqh));
    rc += sizeof(arp_rqh);
    /* Sender MAC address. */
    memcpy(&packet[rc], my_mac, 6);
    rc += 6;
    /* Sender IP address. */
    memcpy(&packet[rc], my_ip, 4);
    rc += 4;
    /* Target MAC address. */
    memset(&packet[rc], 0x00, 6);
    rc += 6;
    /* Target IP address. */
    memcpy(&packet[rc], rem_ip, 4);
    rc += 4;

    return rc;
}

int AssembleArpReply(void)
{
    if (memcmp(packet + 12, &arp_rqh, sizeof(arp_rqh))) {
        printf("No ARP request.\n");
        return -1;
    }
    if (memcmp(packet + 38, my_ip, 4)) {
        printf("Not my IP.\n");
        return -1;
    }
    /* Move source's to destination's MAC address. */
    memcpy(packet, packet + 6, 6);
    /* Source MAC address. */
    memcpy(packet + 6, my_mac, 6);
    /* Set ARP reply opcode. */
    packet[21] = 2;
    /* Move sender's to target's MAC and IP address. */
    memcpy(packet + 32, packet + 22, 10);
    /* Sender MAC address. */
    memcpy(packet + 22, my_mac, 6);
    /* Sender IP address. */
    memcpy(packet + 28, my_ip, 4);

    return 0;
}

typedef struct {
    unsigned char ip_vhl;   /*!< \brief Version and header length. */
    unsigned char ip_tos;   /*!< \brief Type of service. */
    unsigned short ip_len;  /*!< \brief Total length. */
    unsigned short ip_id;   /*!< \brief Identification. */
    unsigned short ip_off;  /*!< \brief Fragment offset field. */
    unsigned char ip_ttl;   /*!< \brief Time to live. */
    unsigned char ip_p;     /*!< \brief Protocol. */
    unsigned short ip_sum;  /*!< \brief Checksum. */
    unsigned long ip_src;   /*!< \brief Source IP address. */
    unsigned long ip_dst;   /*!< \brief Destination IP address. */
} IPHDR;

typedef struct {
    unsigned char  icmp_type;      /*!< \brief Type of message. */
    unsigned char  icmp_code;      /*!< \brief Type sub code. */
    unsigned short icmp_cksum;     /*!< \brief Ones complement header checksum. */
    unsigned long  icmp_spec;      /*!< \brief Type specific information. */
} ICMPHDR;

unsigned short IpChkSumPartial(unsigned short ics, const void *buf, int len)
{
    register unsigned long sum = ics;
    register unsigned short *wp = (unsigned short *) buf;

    /* Sum up 16 bit values. */
    while (len > 1) {
        sum += *wp++;
        len -= 2;
    }

    /* Add remaining byte on odd lengths. */
    if (len) {
        sum += *(unsigned char *) wp;
    }

    /* Fold upper 16 bits to lower ones. */
    while (sum >> 16) {
        sum = (unsigned short)sum + (sum >> 16);
    }
    return (unsigned short) sum;
}



/*!
 * \brief Calculates an the final IP checksum over a block of data.
 *
 * Unlike the partial checksum in NutIpChkSumPartial(), this function takes
 * the one's complement of the final result, thus making it the full checksum.
 */
unsigned short IpChkSum(unsigned short ics, const void *buf, int len)
{
    return ~IpChkSumPartial(ics, buf, len);
}

int AssembleIcmpEchoReply(void)
{
    IPHDR *ip;
    ICMPHDR *icmp;
    unsigned short csum;
    unsigned short dlen;

    if (packet[12] != 0x08 || packet[13] != 0x00) {
        printf("No IP.\n");
        return -1;
    }

    ip = (IPHDR *)(packet + 14);
    if (ip->ip_vhl != 0x45) {
        printf("Not version 4.\n");
        return -1;
    }
    if (ip->ip_p != 0x01) {
        printf("Not ICMP.\n");
        return -1;
    }
    if (memcmp(&ip->ip_dst, my_ip, 4)) {
        printf("Not my IP.\n");
        return -1;
    }

    icmp = (ICMPHDR *)(packet + 34);
    if (icmp->icmp_type != 8) {
        printf("No echo request.\n");
        return -1;
    }

    /* Move source's to destination's MAC address. */
    memcpy(packet, packet + 6, 6);
    /* Source MAC address. */
    memcpy(packet + 6, my_mac, 6);

    /* Set IP destination and source. */
    ip->ip_dst = ip->ip_src;
    memcpy(&ip->ip_src, my_ip, 4);

    /* Set ICMP type. */
    icmp->icmp_type = 0;

    dlen = (ip->ip_len >> 8);
    dlen += (ip->ip_len << 8);
    dlen -= sizeof(IPHDR) + sizeof(ICMPHDR);

    icmp->icmp_cksum = 0;
    csum = IpChkSumPartial(0, icmp, sizeof(ICMPHDR));
    icmp->icmp_cksum = IpChkSum(csum, packet + 42, dlen);

    return 0;
}

char *rn[4][16] = {
    { "CR", "CLDA0", "CLDA1", "BNRY", "TSR", "NCR", NULL, "ISR", "CRDA0", "CRDA1", "ID0", "ID1", "RSR", "CNTR0", "CNTR1", "CNTR2" },
    { "CR", "PAR0", "PAR1", "PAR2", "PAR3", "PAR4", "PAR5", "CURR", "MAR0", "MAR1", "MAR2", "MAR3", "MAR4", "MAR5", "MAR6", "MAR7" },
    { "CR", "PSTART", "PSTOP", NULL, "TPSR", NULL, NULL, NULL, NULL, NULL, NULL, NULL, "RCR", "TCR", "DCR", "IMR" },
    { "CR", "9346CR", "BPAGE", "CONFIG0", "CONFIG1", "CONFIG2", "CONFIG3", "TEST", "CSNSAV", NULL, NULL, "INTR", NULL, "CONFIG4", NULL, NULL }
};

void DumpRegs(void)
{
    unsigned char i;
    unsigned char j;
    volatile unsigned char *rp = (unsigned char *)base;
    unsigned char cr = rp[0];

    for (i = 0; i < 16; i++) {
        printf("%02XH ", i);
        for (j = 0; j < 4; j++) {
            if (i == 0) {
                if ((cr >> 6) == j) {
                    printf("%7s %02XH  ", rn[j][0], cr);
                }
                else {
                    printf("             ");
                }
            }
            else {
                rp[0] = (cr & 0x3F) | (j << 6);
                if (rn[j][i]) {
                    printf("%7s %02XH  ", rn[j][i], rp[i]);
                }
                else {
                    printf("             ");
                }
            }
        }
        puts("");
    }
    rp[0] = cr;
}

void DumpTxStatus(void)
{
    unsigned char tsr;
    unsigned char cnt;

    tsr = nic_inlb(base + NIC_PG0_TSR);
    printf("Tx[0x%02X]:", tsr);
    if (tsr & NIC_TSR_PTX) {
        printf(" OK");
    }
    if (tsr & NIC_TSR_COL) {
        printf(" COL");
    }
    if (tsr & NIC_TSR_ABT) {
        printf(" ABT");
    }
    if (tsr & NIC_TSR_CRS) {
        printf(" CRS");
    }
    if (tsr & NIC_TSR_FU) {
        printf(" FU");
    }
    if (tsr & NIC_TSR_CDH) {
        printf(" CDH");
    }
    if (tsr & NIC_TSR_OWC) {
        printf(" OWC");
    }
    if ((cnt = nic_inlb(base + NIC_PG0_NCR)) != 0) {
        printf(" COL=%u", cnt);
    }
    puts("");
}

void DumpRxStatus(void)
{
    unsigned char rsr;
    unsigned char cnt;

    rsr = nic_inlb(base + NIC_PG0_RSR);
    printf("Rx[0x%02X]:", rsr);
    if (rsr & NIC_RSR_PRX) {
        printf(" OK");
    }
    if (rsr & NIC_RSR_CRC) {
        printf(" CRC");
    }
    if (rsr & NIC_RSR_FAE) {
        printf(" FAE");
    }
    if (rsr & NIC_RSR_FO) {
        printf(" FO");
    }
    if (rsr & NIC_RSR_MPA) {
        printf(" MPA");
    }
    if (rsr & NIC_RSR_PHY) {
        printf(" PHY");
    }
    if (rsr & NIC_RSR_DIS) {
        printf(" DIS");
    }
    if (rsr & NIC_RSR_DFR) {
        printf(" DFR");
    }
    if ((cnt = nic_inlb(base + NIC_PG0_CNTR0)) != 0) {
        printf(" FAE=%u", cnt);
    }
    if ((cnt = nic_inlb(base + NIC_PG0_CNTR1)) != 0) {
        printf(" CRC=%u", cnt);
    }
    if ((cnt = nic_inlb(base + NIC_PG0_CNTR2)) != 0) {
        printf(" Lost=%u", cnt);
    }
    puts("");
}

int NicDiscover(unsigned int start, unsigned int expect)
{
    int rc = 0;
    unsigned int addr;

    base = 0;
    for (addr = start; addr <= 0xFFF0; addr++) {
        if (*((char *)addr) == 0x50 && *((char *)(addr + 1)) == 0x70) {
            if (rc) {
                if (addr == expect + NIC_PG0_8019ID0) {
                    base = addr;
                }
                else if ((base & 0x000F) != NIC_PG0_8019ID0) {
                    base = addr;
                }
            }
            else {
                base = addr;
            }
            rc++;
        }
    }
    if (rc) {
        base -= NIC_PG0_8019ID0;
    }
    return rc;
}

int NicJumperConfig(void)
{
    unsigned char cr;

    /* Select page 3. */
    cr = nic_inlb(base + NIC_CR);
    nic_outlb(base + NIC_CR, NIC_CR_PG3 | (cr & ~(NIC_CR_PS0 | NIC_CR_PS1)));
    /* Check if in jumper mode. */
    if ((nic_inlb(base + NIC_PG3_CONFIG0) & 0x08) == 0) {
        return -1;
    }
    /* Make sure we got the correct base address and IRQ number. */
    if (nic_inlb(base + NIC_PG3_CONFIG1) != 0x80) {
        return -1;
    }
    /* Restore command register. */
    nic_outlb(base + NIC_CR, cr);

    return 0;
}

/*!
 * \brief Detect EEPROM emulation.
 *
 * This routine checks, if EEPROM emulation is available on address
 * bits 13 and 14.
 *
 * Up to board revision 1.3F Ethernut uses a pull up resistor on pin 
 * EEDO, which sets all bits of CONFIG3 to 1. Unfortunately we are not 
 * able to modify the full duplex bit. The only solution is to use a 
 * real EEPROM or emulate one. 
 *
 * \return 0 if EEPROM emulation has been detected, otherwise returns -1.
 */
int NicEepromDetect(void)
{
    unsigned int cnt = 0;
    unsigned short tmo;
    unsigned char cr;

    /* Select page 3. */
    cr = nic_inlb(base + NIC_CR);
    nic_outlb(base + NIC_CR, NIC_CR_PG3 | (cr & ~(NIC_CR_PS0 | NIC_CR_PS1)));
    nic_outlb(base + NIC_PG3_EECR, NIC_EECR_EEM0);

    /* Disable external memory interface. */
    PORTC = 0x40;
    DDRC = 0xC0;
    MCUCR &= ~_BV(SRE);

    do {
        tmo = 0;
        if(PINC & _BV(5)) {
            cnt++;
            while(++tmo && (PINC & 0x20));
        }
        else {
            while(++tmo && (PINC & 0x20) == 0);
        }
    } while (tmo);

    /* Re-enable external memory interface. */
    MCUCR |= _BV(SRE);
    PORTC = 0x00;
    DDRC = 0x00;

    /* Restore command register. */
    nic_outlb(base + NIC_CR, cr);

    /* 32 clock cycles are generated for each of the 9 data words. */
    return cnt == 288 ? 0 : -1;
}

/*!
 * \brief Emulate Ethernet Controller EEPROM.
 *
 * On Ethernut 1.3 Rev. G address line A14 (port C, bit 6) is connected
 * to EEDO and address line A13 (port C, bit 5) is connected to EESK.
 * Switching off the external memory bus allows to simulate an EEPROM
 * by software.
 *
 * \return 0 on success, -1 otherwise.
 */
int NicEepromEmulate(void)
{
    register unsigned char clk;
    register unsigned char val;
    volatile unsigned char tmo;

    PORTC = 0x40;
    DDRC = 0xC0;

    nic_outlb(base + NIC_CR, NIC_CR_PG3 | NIC_CR_STP | NIC_CR_RD2);
    nic_outlb(base + NIC_PG3_EECR, NIC_EECR_EEM0);

    MCUCR &= ~_BV(SRE);

    /* 
     * Send 9 data words. While running in jumper mode, all register
     * values are ignored except CONFIG3 and CONFIG4. Thus we send
     * 0xFF30 for each data word.
     */
    for(val = 9; val--;) {
        /* 
         * Skip 1 start bit, 3 command bits, 6 address bits and 
         * 1 direction change bit and send high data byte 0xFF.
         */
        for(clk = 19; clk--; ) {
            tmo = 0;
            while(--tmo && bit_is_clear(PINC, 5));
            while(--tmo && bit_is_set(PINC, 5));
        }
        /* Send low data byte 0x30 and skip 5 bits trailer. */
        for(clk = 13; clk--;) {
            tmo = 0;
            while(--tmo && bit_is_clear(PINC, 5));
            if (clk == 9 || clk == 10) {
                PORTC = 0x40;
            }
            else {
                PORTC = 0x00;
            }
            while(--tmo && bit_is_set(PINC, 5));
        }
        PORTC = 0x40;
    }

    MCUCR |= _BV(SRE);
    PORTC = 0x00;
    DDRC = 0x00;

    nic_outlb(base + NIC_CR, NIC_CR_PG3 | NIC_CR_STP | NIC_CR_RD2);
    val = nic_inlb(base + NIC_PG3_CONFIG3);
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STP | NIC_CR_RD2);

    return (val == 0x30) ? 0 : -1;
}

int NicWaitLink(void)
{
    volatile unsigned long tmo = 1000000;
    unsigned char cr;

    /* Get configuration register 0 from page 3. */
    cr = nic_inlb(base + NIC_CR);
    nic_outlb(base + NIC_CR, NIC_CR_PG3 | cr);
    do {
        if ((nic_inlb(base + NIC_PG3_CONFIG0) & 0x24) == 0) {
            break;
        }
    } while (--tmo);
    nic_outlb(base + NIC_CR, cr);

    return tmo ? 0 : -1;
}

void NicPowerDown(void)
{
    /* Select page 3. */
    nic_outlb(base + NIC_CR, NIC_CR_PG3 | NIC_CR_STP | NIC_CR_RD2);

    /*
     * The nic configuration registers are write protected unless both
     * EEM bits are set to 1.
     */
    nic_outlb(base + NIC_PG3_EECR, NIC_EECR_EEM0 | NIC_EECR_EEM1);

    /* 
     * LED0: Yellow, lit when output low. 
     * LED1: Green, lit when output high.
     */
    nic_outlb(base + NIC_PG3_CONFIG3, 0x20);
    Delay(200000);
    nic_outlb(base + NIC_PG3_CONFIG3, nic_inlb(NIC_PG3_CONFIG3) | NIC_CONFIG3_PWRDN);
    nic_outlb(base + NIC_PG3_HLTCLK, 'H');

}

void NicPowerUp(void)
{
    nic_outlb(base + NIC_PG3_HLTCLK, 'R');
    Delay(2000);
    nic_outlb(base + NIC_PG3_CONFIG3, nic_inlb(NIC_PG3_CONFIG3) & ~(NIC_CONFIG3_SLEEP | NIC_CONFIG3_PWRDN));
    //Delay(20000);

    /*
     * Reenable write protection of the nic configuration registers
     * and wait for link test to complete.
     */
    nic_outlb(base + NIC_PG3_EECR, 0);

    /* Select page 3. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STP | NIC_CR_RD2);
}

int NicWaitReset(void)
{
    volatile unsigned short tmo = 0;

    while (--tmo && (nic_inlb(base + NIC_PG0_ISR) & NIC_ISR_RST) == 0);
    nic_outlb(base + NIC_PG0_RCR, NIC_RCR_MON);

    return tmo ? 0 : -1;
}

int NicStop(void)
{
    if ((nic_inlb(base + NIC_PG0_ISR) & NIC_ISR_RST) == 0) {
        nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STP | NIC_CR_RD2);
        return NicWaitReset();
    }
    return 0;
}

int NicReset(void)
{
    NicStop();
    nic_outlb(base + NIC_RESET, nic_inlb(base + NIC_RESET));
    return NicWaitReset();
}

int NicPromiscuous(void)
{
    unsigned char i;

    NicStop();

    /* Select page 1. */
    nic_outlb(base + NIC_CR, NIC_CR_PG1 | NIC_CR_STP | NIC_CR_RD2);

    for (i = 0; i < 8; i++) {
        nic_outlb(base + NIC_PG1_MAR0 + i, 0xFF);
    }
    /* Select page 0. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STP | NIC_CR_RD2);
    /* Initialize receive configuration register. */
    nic_outlb(base + NIC_PG0_RCR, NIC_RCR_AB | NIC_RCR_AM | NIC_RCR_PRO);

    /* Put NIC in start mode. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STA | NIC_CR_RD2);
    
    return 0;
}

int NicStart(void)
{
    unsigned char i;

    /* Select page 3. */
    nic_outlb(base + NIC_CR, NIC_CR_PG3 | NIC_CR_STP | NIC_CR_RD2);

    /*
     * The nic configuration registers are write protected unless both
     * EEM bits are set to 1.
     */
    nic_outlb(base + NIC_PG3_EECR, NIC_EECR_EEM0 | NIC_EECR_EEM1);

    /*
     * Network media had been set to 10Base2 by the virtual EEPROM and
     * will be set now to auto detect. This will initiate a link test.
     * We don't force 10BaseT, because this would disable the link test.
     */
    nic_outlb(base + NIC_PG3_CONFIG2, NIC_CONFIG2_BSELB);

    /*
     * Disable sleep and power down.
     */
    nic_outlb(base + NIC_PG3_CONFIG3, 0);

    /*
     * Reenable write protection of the nic configuration registers
     * and wait for link test to complete.
     */
    nic_outlb(base + NIC_PG3_EECR, 0);
    Delay(200000);

    /* Select page 0. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STP | NIC_CR_RD2);
    /* 
     * Initialize data configuration register to byte transfer,
     * normal mode and FIFO threshold of 8 bytes.
     */
    nic_outlb(base + NIC_PG0_DCR, NIC_DCR_FT1 | NIC_DCR_LS);
    /* Clear remote byte count registers. */
    nic_outlb(base + NIC_PG0_RBCR0, 0);
    nic_outlb(base + NIC_PG0_RBCR1, 0);
    /* Initialize receive configuration register. */
    nic_outlb(base + NIC_PG0_RCR, NIC_RCR_MON);
    /* Place NIC in internal loopback mode (mode 1). */
    nic_outlb(base + NIC_PG0_TCR, NIC_TCR_LB0);
    /* Initialize receive buffer ring. */
    nic_outlb(base + NIC_PG0_TPSR, 0x40);
    nic_outlb(base + NIC_PG0_BNRY, 0x60 - 1);
    nic_outlb(base + NIC_PG0_PSTART, 0x46);
    nic_outlb(base + NIC_PG0_PSTOP, 0x60);
    /* Clear interrupt status register. */
    nic_outlb(base + NIC_PG0_ISR, 0xFF);
    /* Initialize interrupt mask register. */
    nic_outlb(base + NIC_PG0_IMR, 0x00);
    /* Select page 1. */
    nic_outlb(base + NIC_CR, NIC_CR_PG1 | NIC_CR_STP | NIC_CR_RD2);
    /* Initialize physical address register. */
    for (i = 0; i < 6; i++) {
        nic_outlb(base + NIC_PG1_PAR0 + i, my_mac[i]);
    }
    /* Initialize multicast address register. */
    for (i = 0; i < 8; i++) {
        //nic_outlb(base + NIC_PG1_MAR0 + i, 0x00);
        nic_outlb(base + NIC_PG1_MAR0 + i, 0xFF);
    }
    /* Initialize current pointer. */
    nic_outlb(base + NIC_PG1_CURR, 0x46);
    /* Switch back to register page 0, remaining in stop mode. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STP | NIC_CR_RD2);

    /*
     * Take receiver out of monitor mode and enable it for accepting
     * broadcasts.
     */
    //nic_outlb(base + NIC_PG0_RCR, NIC_RCR_AB);
    nic_outlb(base + NIC_PG0_RCR, NIC_RCR_AB | NIC_RCR_AM | NIC_RCR_PRO);

    /* Put NIC in start mode. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STA | NIC_CR_RD2);
    /* Switch transmit configuration register from loopback to normal mode. */
    nic_outlb(base + NIC_PG0_TCR, 0x00);

    return NicWaitLink();
}

void NicDmaPrepare(unsigned int addr, int len)
{
    /* Reset remote dma complete flag. */
    nic_outlb(base + NIC_PG0_ISR, NIC_ISR_RDC);
    /* Set remote dma byte count and start address. */
    nic_outlb(base + NIC_PG0_RBCR0, len);
    nic_outlb(base + NIC_PG0_RBCR1, len >> 8);
    nic_outlb(base + NIC_PG0_RSAR0, addr);
    nic_outlb(base + NIC_PG0_RSAR1, addr >> 8);
}

void NicDmaComplete(unsigned char cr)
{
    volatile unsigned char tmo = 0;

    /* Wait until DMA complete. */
    while (--tmo && (nic_inlb(base + NIC_PG0_ISR) & NIC_ISR_RDC) == 0);
    /* Complete remote DMA. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_RD2 | cr);
}

void NicDmaRead(unsigned int addr, unsigned char *data, int len)
{
    unsigned char cr = nic_inlb(base + NIC_CR) & (NIC_CR_STP | NIC_CR_STA);

    NicDmaPrepare(addr, len);
    /* Set read command. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_RD0 | cr);
    /* Transfer the data. */
    while (len--) {
        *data++ = nic_inlb(base + NIC_IOPORT);
    }
    NicDmaComplete(cr);
}

void NicDmaWrite(unsigned int addr, const unsigned char *data, int hlen, unsigned char pad, int len)
{
    unsigned char cr = nic_inlb(base + NIC_CR) & (NIC_CR_STP | NIC_CR_STA);

    NicDmaPrepare(addr, len);
    /* Set write command. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_RD1 | cr);
    len -= hlen;
    if (len < 0) {
        len = 0;
    }
    /* Transfer the data. */
    while (hlen--) {
        nic_outlb(base + NIC_IOPORT, *data);
        data++;
    }
    /* Transfer padding bytes. */
    while (len--) {
        nic_outlb(base + NIC_IOPORT, pad);
    }
    NicDmaComplete(cr);
}

int NicMemTestPattern(unsigned char pattern)
{
    unsigned int addr;
    unsigned char val[16];
    unsigned char i;

    NicDmaWrite(0x4000, 0, 0, pattern, 8192);
    for (addr = 0x4000; addr < 0x6000; addr += sizeof(val)) {
        memset(val, ~pattern, sizeof(val));
        NicDmaRead(addr, val, sizeof(val));
        for (i = 0; i < sizeof(val); i++) {
            if (val[i] != pattern) {
                return -1;
            }
        }
    }
    return 0;
}

int NicMemTest(void)
{
    unsigned char pattern[] = { 0x00, 0xFF, 0xAA, 0x55 };
    unsigned char rval[32];
    int i;
    int j;
    int k;

    /* Pattern test. */
    for (i = 0; i < sizeof(pattern); i++) {
        if (NicMemTestPattern(pattern[i])) {
            return -1;
        }
    }

    /* Walking bit test. */
    for (i = 0x80; i; i >>= 1) {
        if (NicMemTestPattern((unsigned char)i)) {
            return -1;
        }
    }

    /* Address test. */
    NicDmaWrite(0x4000, 0, 0, 0x00, 8192);
    for (i = 0x4000; i < 0x6000; i += sizeof(rval)) {
        NicDmaWrite(i, 0, 0, 0xFF, sizeof(rval));
        for (j = 0x4000; j < 0x6000; j += sizeof(rval)) {
            if (j != i) {
                NicDmaRead(j, rval, sizeof(rval));
                for (k = 0; k < sizeof(rval); k++) {
                    if (rval[k]) {
                        return -1;
                    }
                }
            }
        }
        NicDmaWrite(i, 0, 0, 0x00, sizeof(rval));
    }
    return 0;
}

int NicReceive(unsigned char *buf, int siz)
{
    int rc = 0;
    unsigned char isr;
    unsigned char curr;
    unsigned char bnry;
    unsigned char nextpg;
    NIC_PKT_HDR hdr;

    isr = nic_inlb(base + NIC_PG0_ISR);
    nic_outlb(base + NIC_PG0_ISR, isr);
    if (isr & NIC_ISR_PRX) {
        nic_outlb(base + NIC_CR, NIC_CR_PG1 | NIC_CR_STA | NIC_CR_RD2);
        //Delay(16);
        curr = nic_inlb(base + NIC_PG1_CURR);
        nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STA | NIC_CR_RD2);

        if ((bnry = nic_inlb(base + NIC_PG0_BNRY) + 1) >= NIC_STOP_PAGE)
            bnry = NIC_FIRST_RX_PAGE;

        if (bnry == curr) {
            return 0;
        }

        /* Read the NIC specific packet header. */
        NicDmaRead(bnry << 8, (unsigned char *) & hdr, sizeof(NIC_PKT_HDR));
        rc = hdr.ph_size - 4;
        if (rc > siz) {
            rc = siz;
        }

        /*
         * Calculate the page of the next packet. If it differs from the
         * pointer in the packet header, we return with errorcode.
         */
        nextpg = bnry + (hdr.ph_size >> 8) + ((hdr.ph_size & 0xFF) != 0);
        if (nextpg >= NIC_STOP_PAGE) {
            nextpg -= NIC_STOP_PAGE;
            nextpg += NIC_FIRST_RX_PAGE;
        }
        if (nextpg != hdr.ph_nextpg) {
            unsigned char nextpg1 = nextpg + 1;
            if (nextpg1 >= NIC_STOP_PAGE) {
                nextpg1 -= NIC_STOP_PAGE;
                nextpg1 += NIC_FIRST_RX_PAGE;
            }
            if (nextpg1 != hdr.ph_nextpg) {
                return 0;
            }
            nextpg = nextpg1;
        }
        if ((hdr.ph_status & 0x0E) == 0 && siz) {
            NicDmaRead((bnry << 8) + sizeof(NIC_PKT_HDR), buf, rc);
        }

        /* Set boundary register to the last page we read. */
        if (--nextpg < NIC_FIRST_RX_PAGE) {
            nextpg = NIC_STOP_PAGE - 1;
        }
        nic_outlb(base + NIC_PG0_BNRY, nextpg);
    }
    return rc;
}

int NicTransmitWait(void)
{
    volatile unsigned long tmo = 1000000;
    while ((nic_inlb(base + NIC_CR) & NIC_CR_TXP) != 0 && --tmo);

    return tmo ? 0 : -1;
}

unsigned long CalcCrc32(int hlen, int len)
{
    unsigned long crc = 0xFFFFFFFFUL;
    unsigned long pol = 0xEDB88320UL;
    unsigned char oct;
    int i;
    unsigned char j;

    for(i = 0; i < len; i++){
        oct = (i < hlen) ? packet[i] : 0;
        for(j = 8; j--; oct >>= 1) {
            if((crc ^ oct) & 1) {
                crc >>= 1;
                crc ^= pol;
            }
            else {
                crc >>= 1;
            }
        }
    }
    return crc ^ 0xFFFFFFFFUL;
}

void NicTransmit(int hlen, int len)
{
    NicTransmitWait();
    NicDmaWrite(NIC_FIRST_TX_PAGE << 8, packet, hlen, 0x00, len);

    /* Number of bytes to be transmitted. */
    nic_outlb(base + NIC_PG0_TBCR0, (unsigned char)len);
    nic_outlb(base + NIC_PG0_TBCR1, (unsigned char)(len >> 8));

    /* First page of packet to be transmitted. */
    nic_outlb(base + NIC_PG0_TPSR, NIC_FIRST_TX_PAGE);

    /* Start transmission. */
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STA | NIC_CR_TXP | NIC_CR_RD2);
}

int NicIrqCheck(int hlen)
{
    unsigned char msk = 0x60;
    unsigned char st;

    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STP | NIC_CR_RD2);

    nic_outlb(base + NIC_PG0_IMR, 0x00);
    nic_outlb(base + NIC_PG0_ISR, 0xFF);
    DDRE &= ~msk;
    PORTE &= ~msk;
    Delay(200);
    st = PINE & msk;
    if (st) {
        nic_outlb(base + NIC_PG0_IMR, 0x7F);

        NicTransmit(hlen, 1480);
        NicTransmitWait();
        st ^= (PINE & msk);
        if (st) {
            nic_outlb(base + NIC_PG0_IMR, 0x00);
            nic_outlb(base + NIC_PG0_ISR, 0xFF);
            Delay(200);
            st &= PINE;
        }
    }
    return st ? 0 : -1;
}

int NicWriteCheck(void)
{
    int i;

    nic_outlb(base + NIC_CR, NIC_CR_PG1 | NIC_CR_STP | NIC_CR_RD2);
    if ((nic_inlb(base + NIC_CR) & (NIC_CR_PS0 | NIC_CR_PS1)) != NIC_CR_PG1) {
        return -1;
    }
    /* Initialize physical address register. */
    for (i = 0; i < 6; i++) {
        nic_outlb(base + NIC_PG1_PAR0 + i, my_mac[i]);
    }
    /* Initialize multicast address register. */
    for (i = 0; i < 8; i++) {
        nic_outlb(base + NIC_PG1_MAR0 + i, 0x00);
    }
    /* Verify physical address register. */
    for (i = 0; i < 6; i++) {
        if (nic_inlb(base + NIC_PG1_PAR0 + i) != my_mac[i]) {
            return -1;
        }
    }
    /* Verify multicast address register. */
    for (i = 0; i < 8; i++) {
        if (nic_inlb(base + NIC_PG1_MAR0 + i)) {
            return -1;
        }
    }
    nic_outlb(base + NIC_CR, NIC_CR_PG0 | NIC_CR_STP | NIC_CR_RD2);

    return 0;
}

int NicLoopback(int hlen, int len, unsigned char mode)
{
    int i;
    unsigned long fcs;
    unsigned long got;

    nic_outlb(base + NIC_PG0_RCR, 0);
    nic_outlb(base + NIC_PG0_DCR, NIC_DCR_FT1);
    nic_outlb(base + NIC_PG0_TCR, mode);
    nic_outlb(base + NIC_PG0_ISR, nic_inlb(base + NIC_PG0_ISR));
    NicTransmit(hlen, len);
    NicTransmitWait();

    /* First byte in FIFO must be a trailer byte. */
    if (nic_inlb(base + NIC_PG0_FIFO)) {
        return -1;
    }
    /* Next four bytes are the FCS. */
    fcs = 0;
    for (i = 0; i < 4; i++) {
        fcs |= (unsigned long)nic_inlb(base + NIC_PG0_FIFO) << (i * 8);
    }
    if (CalcCrc32(hlen, len) != fcs) {
        return -1;
    }
    /* Last two bytes are the number of bytes received. */
    got = nic_inlb(base + NIC_PG0_FIFO);
    got |= (unsigned long)nic_inlb(base + NIC_PG0_FIFO) << 8;
    if (got != len + 4) {
        return -1;
    }
    return 0;
}

int NicTest(void)
{
    int hlen;

    if (NicDiscover(0x1100, 0x8300) == 0) {
        printf("404 No controller detected.\n");
        return -1;
    }
    if (NicWriteCheck()) {
        printf("401 NIC write failed.\n");
        return -1;
    }
    if (NicJumperConfig()) {
        printf("404 Bad jumper setting.\n");
        return -1;
    }
    if (base != 0x8300) {
        printf("404 Bad base address %X.\n", base);
        return -1;
    }
    if (NicEepromDetect() || NicEepromEmulate()) {
        printf("401 EEPROM emulation failed.\n");
        return -1;
    }
    if (NicDiscover(0x8000, 0x8300) != 1) {
        printf("401 Address decode error.\n");
        return -1;
    }
    if (NicStop()) {
        printf("401 NIC stop failed.\n");
        return -1;
    }
    if (NicReset()) {
        printf("401 NIC reset failed.\n");
        return -1;
    }
    if (NicMemTest()) {
        printf("401 NIC memory test failed.\n");
        return -1;
    }
    hlen = AssembleArpRequest(0);
    if (NicStart()) {
        printf("401 Link failed.\n");
        return -1;
    }
    if (NicLoopback(hlen, 12 + 32 + 5, NIC_TCR_LB0)) {
        printf("421 Loopback 1 failed.\n");
        return -1;
    }
    if (NicReset()) {
        printf("401 NIC reset failed.\n");
        return -1;
    }
    Delay(200000);
    if (NicStart()) {
        printf("401 Link failed.\n");
        return -1;
    }
    if (NicLoopback(hlen, 12 + 48 + 5, NIC_TCR_LB1)) {
        printf("422 Loopback 2 failed.\n");
        return -1;
    }
    if (NicReset()) {
        printf("401 NIC reset failed.\n");
        return -1;
    }
    Delay(200000);
    if (NicStart()) {
        printf("401 Link failed.\n");
        return -1;
    }
    if (NicLoopback(hlen, 12 + 1480 + 5, NIC_TCR_LB1 | NIC_TCR_LB0)) {
        printf("423 Loopback 3 failed.\n");
        return -1;
    }
    if (NicReset()) {
        printf("401 NIC reset failed.\n");
        return -1;
    }
    Delay(200000);
    if (NicStart()) {
        printf("401 Link failed.\n");
        return -1;
    }
    if (NicIrqCheck(hlen)) {
        printf("430 IRQ failed\n");
        return -1;
    }
    return 0;
}

int main(void)
{
    int hlen;
    int got;

    cli();
    MCUCR |= _BV(SRE);
    StdIoInit();

    printf("\n\nRTL8019AS Test\n");

    if (NicTest()) {
        DumpRegs();
        for (;;);
    }
    printf("All tests passed.\n");

    printf("Ready for ping %u.%u.%u.%u (MAC %02X:%02X:%02X:%02X:%02X:%02X)\n",
        my_ip[0], my_ip[1], my_ip[2], my_ip[3],
        my_mac[0], my_mac[1], my_mac[2], my_mac[3], my_mac[4], my_mac[5]);

    hlen = AssembleArpRequest(0);
    for(;;) {
        if ((got = NicReceive(packet, sizeof(packet))) > 0) {
            DumpRxStatus();
            if (AssembleArpReply() == 0) {
                puts("ARP");
                NicTransmit(got, got);
            }
            else if (AssembleIcmpEchoReply() == 0) {
                puts("Ping");
                NicTransmit(got, got);
            }
        }
    }
    return 0;
}
